#![allow(clippy::type_complexity)]

use super::asset_handler::AssetHandler;

use super::engineobject::*;
use super::enginesettings::EngineSettings;

use anyhow::Result;
use assetpath::AssetPath;
use context::prelude::cgmath::vec3;
use loading_screen::LoadingScreen;

use crate::assets::assetmanager::AssetManager;
use crate::assets::entity_object_manager::EntityObjectManager;
use crate::prelude::*;

use std::sync::{Arc, RwLock, RwLockReadGuard};
use std::{collections::HashMap, marker::PhantomData};
use std::{
    ptr::NonNull,
    sync::{Mutex, MutexGuard},
};

#[derive(Default, Debug, Clone)]
pub struct EngineCreateInfo<'a> {
    pub app_info: ApplicationInfo,

    pub window_info: WindowCreateInfo,

    pub os_specific_config: OsSpecificConfig,

    pub vulkan_debug_info: VulkanDebugInfo,
    pub volume_info: HashMap<String, f32>,

    pub gui_info: GuiHandlerCreateInfo<'a>,

    pub enable_backtrace: bool,
    pub enable_mouse: bool,
    pub enable_keyboard: bool,
    pub enable_controller: bool,

    pub controller_deadzones: ControllerDeadzones,

    pub resource_base_path: String,

    // engine settings
    pub controller_directories: ControllerPictureDirectories,
    pub asset_directories: AssetDirectories,

    pub graphics_info: GraphicsInfo,
    pub raytracing_info: RaytracingInfo,
    pub rasterizer_info: RasterizerInfo,
    pub key_backgrounds: ButtonBackgrounds,
}

struct GuiPostProcess(Arc<GuiHandler>);

impl context::prelude::PostProcess for GuiPostProcess {
    fn priority(&self) -> u32 {
        50
    }

    fn process(
        &self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        indices: &TargetMode<usize>,
    ) -> Result<()> {
        self.0.process(buffer_recorder, indices)
    }

    fn resize(&self, width: u32, height: u32) -> Result<()> {
        self.0.resize(width, height)
    }
}

struct GraphicsObjects {
    context: Arc<Context>,
    gui_handler: Arc<GuiHandler>,
    gui_post_process: Arc<GuiPostProcess>,
    input: Arc<RwLock<Input>>,

    // loads and keeps track of raw data
    asset_manager: RwLock<AssetManager>,

    resource_base_path: String,
}

impl Drop for GraphicsObjects {
    fn drop(&mut self) {
        self.context
            .render_core()
            .remove_post_processing_routine(self.gui_post_process.clone());
    }
}

pub struct Engine {
    engine_settings: Arc<EngineSettings>,

    graphical: GraphicsObjects,

    entity_object_manager: EntityObjectManager,
}

impl Engine {
    pub fn new(#[allow(unused)] mut create_info: EngineCreateInfo<'_>) -> Result<Arc<Self>> {
        if let Font::Path(path) = &mut create_info.gui_info.font {
            path.set_prefix(&create_info.resource_base_path);
        }
        create_info.gui_info.resource_directory.assume_prefix_free();

        create_info
            .controller_directories
            .xbox_path
            .set_prefix(&create_info.resource_base_path);
        create_info
            .controller_directories
            .steam_path
            .set_prefix(&create_info.resource_base_path);
        create_info
            .controller_directories
            .ps4_path
            .set_prefix(&create_info.resource_base_path);

        create_info
            .asset_directories
            .entity_file_directory
            .set_prefix(&create_info.resource_base_path);
        create_info
            .asset_directories
            .gltf_file_directory
            .set_prefix(&create_info.resource_base_path);
        create_info
            .asset_directories
            .tile_file_directory
            .set_prefix(&create_info.resource_base_path);

        create_info
            .key_backgrounds
            .light
            .set_prefix(&create_info.resource_base_path);
        create_info
            .key_backgrounds
            .dark
            .set_prefix(&create_info.resource_base_path);

        // ---------------  context  --------------------

        let context = {
            create_info.app_info.engine_name = "EngineNameNeeded".to_string();
            create_info.app_info.engine_version = 1;

            // gui handler needs color attachment
            // rest is depending on render type
            let image_usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;

            let device_features = DeviceFeatures {
                samplerAnisotropy: VK_TRUE,

                ..Default::default()
            };

            let mut context_builder = Context::new()
                .set_volume_info(create_info.volume_info)
                .set_app_info(create_info.app_info)
                .set_window_info(create_info.window_info)
                .set_os_specific_info(create_info.os_specific_config)
                .set_vulkan_debug_info(create_info.vulkan_debug_info)
                .set_controller_deadzones(create_info.controller_deadzones)
                .set_sample_count(create_info.graphics_info.sample_count)
                .set_render_core_info(
                    VK_FORMAT_R8G8B8A8_SRGB,
                    image_usage,
                    create_info.graphics_info.vsync,
                )
                .set_device_features(device_features);

            if create_info.enable_mouse {
                context_builder = context_builder.enable_mouse();
            }

            if create_info.enable_keyboard {
                context_builder = context_builder.enable_keyboard();
            }

            if create_info.enable_controller {
                context_builder = context_builder.enable_controller();
            }

            if let SceneType::Raytracer = create_info.graphics_info.render_type {
                context_builder = context_builder.enable_ray_tracing();
            }

            if create_info.enable_backtrace {
                context_builder = context_builder.enable_backtrace();
            }

            context_builder.build(|device, queue, window_extents, images| {
                Scene::new(
                    device,
                    queue,
                    window_extents,
                    images,
                    create_info.rasterizer_info,
                    create_info.raytracing_info,
                    create_info.graphics_info,
                )
            })?
        };

        // update sound handler center
        {
            let sound_handler = context.sound();
            sound_handler.set_direction(vec3(0.0, 1.0, 0.0), vec3(0.0, 0.0, 1.0))?;
        }

        // ---------------  settings  -------------------

        {
            create_info.graphics_info.render_type = Scene::check_scene_type(
                context.device(),
                create_info.graphics_info.render_type,
                create_info.rasterizer_info.use_deferred,
            )
            .ok_or_else(|| anyhow::Error::msg("No proper render type could be found"))?;
        }

        let engine_settings = Arc::new(EngineSettings::new(
            &context,
            create_info.graphics_info,
            create_info.raytracing_info,
            create_info.rasterizer_info,
            create_info.controller_directories,
            create_info.asset_directories,
            create_info.key_backgrounds,
            create_info.resource_base_path.clone(),
        )?);

        let engine = Arc::new(Engine {
            graphical: {
                let asset_manager = AssetManager::new(engine_settings.clone())?;

                let gui_handler = GuiHandler::new(
                    create_info.gui_info,
                    &(context.clone() as Arc<dyn ContextInterface>),
                )?;

                // default keyboard navigation
                let mut direction_mapping = HashMap::new();
                // direction_mapping.insert(Keycode::A, GuiDirection::Left);
                // direction_mapping.insert(Keycode::D, GuiDirection::Right);
                // direction_mapping.insert(Keycode::W, GuiDirection::Up);
                // direction_mapping.insert(Keycode::S, GuiDirection::Down);
                direction_mapping.insert(Keycode::Left, GuiDirection::Left);
                direction_mapping.insert(Keycode::Right, GuiDirection::Right);
                direction_mapping.insert(Keycode::Up, GuiDirection::Up);
                direction_mapping.insert(Keycode::Down, GuiDirection::Down);

                let input = Input { direction_mapping };

                let gui_post_process = Arc::new(GuiPostProcess(gui_handler.clone()));
                context
                    .render_core()
                    .add_post_processing_routine(gui_post_process.clone());

                GraphicsObjects {
                    context,
                    gui_post_process,
                    gui_handler,
                    input: Arc::new(RwLock::new(input)),

                    asset_manager: RwLock::new(asset_manager),

                    resource_base_path: create_info.resource_base_path,
                }
            },

            entity_object_manager: EntityObjectManager::default(),

            engine_settings,
        });

        Ok(engine)
    }

    fn convert_scene_ref(tscene: &Box<dyn TScene>) -> Option<NonNull<Scene>> {
        unsafe {
            let ptr_to_ptr: *const *mut Scene =
                std::mem::transmute(destructure_traitobject::data(tscene as *const _));

            NonNull::new(*ptr_to_ptr)
        }
    }

    pub fn scene(&self) -> &Scene {
        unsafe {
            Self::convert_scene_ref(self.graphical.context.render_core_mut().scene_mut())
                .unwrap()
                .as_ref()
        }
    }

    pub fn on_scene<'a, F>(&'a self, f: F) -> Result<()>
    where
        F: FnOnce(&Scene) -> Result<()> + 'a,
    {
        f(unsafe {
            Self::convert_scene_ref(self.graphical.context.render_core().scene())
                .unwrap()
                .as_ref()
        })
    }

    pub fn scene_mut(&self) -> &mut Scene {
        unsafe {
            Self::convert_scene_ref(self.graphical.context.render_core_mut().scene_mut())
                .unwrap()
                .as_mut()
        }
    }

    pub fn on_scene_mut<'a, F>(&'a self, f: F) -> Result<()>
    where
        F: FnOnce(&mut Scene) -> Result<()> + 'a,
    {
        f(unsafe {
            Self::convert_scene_ref(self.graphical.context.render_core_mut().scene_mut())
                .unwrap()
                .as_mut()
        })
    }

    pub fn set_game_object<E>(&self, engine_object: Option<E>) -> Result<()>
    where
        E: EngineObject + Send + Sync,
    {
        self.context().set_context_object(match engine_object {
            Some(engine_object) => Some(ContextObjectImpl::new(
                self.graphical.context.clone(),
                engine_object,
                self.graphical.input.clone(),
                self.gui_handler().clone(),
            )?),
            None => None,
        });

        Ok(())
    }

    pub fn loading_screen<T, R, L, G>(
        &self,
        gui: Option<Arc<G>>,
        loader: L,
        on_ready: R,
    ) -> Result<()>
    where
        R: Fn(T) -> anyhow::Result<()> + 'static + Send + Sync,
        T: Send + Sync + 'static,
        L: FnOnce() -> T + Send + Sync + 'static,
        G: TopLevelGui + TopGui + Send + Sync + 'static,
    {
        LoadingScreen::load(self.context(), gui, loader, on_ready)
    }

    pub fn set_fallback<F>(&self, fallback: F)
    where
        F: Fn(anyhow::Error) -> anyhow::Result<()> + 'static + Send + Sync,
    {
        self.context().set_fallback(fallback);
    }

    pub fn run(self: &Arc<Engine>) -> Result<()> {
        self.graphical.context.run()
    }

    pub fn new_point_light(&self) -> Result<Light> {
        Light::point_light(self.context().device())
    }

    pub fn new_directional_light(&self) -> Result<Light> {
        Light::directional_light(self.context().device())
    }

    pub fn new_spot_light(&self) -> Result<Light> {
        Light::spot_light(self.context().device())
    }

    pub fn window_config(&self) -> WindowConfig<'_> {
        self.context().window_config()
    }

    pub fn quit(&self) -> Result<()> {
        self.context().close()
    }

    pub fn gui_handler(&self) -> &Arc<GuiHandler> {
        &self.graphical.gui_handler
    }

    pub fn device(&self) -> &Arc<Device> {
        self.context().device()
    }

    pub fn queue(&self) -> &Arc<Mutex<Queue>> {
        self.context().queue()
    }

    pub fn sound(&self) -> MutexGuard<'_, SoundHandler> {
        self.context().sound()
    }

    pub fn controllers(&self) -> RwLockReadGuard<'_, Vec<Arc<RwLock<Controller>>>> {
        self.context().controllers()
    }

    pub fn active_controller(&self) -> Result<Option<Arc<RwLock<Controller>>>> {
        self.context().active_controller()
    }

    pub fn set_active_controller(&self, controller: &Arc<RwLock<Controller>>) {
        self.context().set_active_controller(controller);
    }

    pub fn context(&self) -> &Arc<Context> {
        &self.graphical.context
    }

    pub fn build_path(&self, path: &str) -> AssetPath {
        (self.graphical.resource_base_path.as_str(), path).into()
    }

    pub fn assets(&self) -> AssetHandler<'_> {
        AssetHandler {
            asset_manager: &self.graphical.asset_manager,

            context: &self.graphical.context,

            resource_base_path: &self.graphical.resource_base_path,

            entity_object_manager: &self.entity_object_manager,

            phantom_data: PhantomData,
        }
    }
}

// getter
impl Engine {
    pub fn settings(&self) -> &Arc<EngineSettings> {
        &self.engine_settings
    }
}
