use crate::prelude::*;

use core::fmt;
use std::any::TypeId;
use std::cmp::Ordering;
use std::{num::ParseIntError, str::FromStr};

use anyhow::Result;

pub const COMPONENT_SEPARATOR: char = '§';

// Used for serialisation over network
#[derive(Serialize, Deserialize)]
pub struct EntityCreateInfo {
    gltf_file: Option<String>,
    network_id: Option<NetworkID>,
    ignore_transactions: bool,
}

impl ComponentCreateInfo<'_> for EntityCreateInfo {}

#[derive(Default)]
pub struct ActivationState {
    activated: bool,
}

impl ActivationState {
    pub fn is_active(&self) -> bool {
        self.activated
    }

    pub(crate) fn apply_change(&mut self) {
        self.activated = !self.activated;
    }
}

pub struct EntityObject {
    #[cfg(debug_assertions)]
    pub debug_name: Option<String>,

    // gltf file name
    pub(crate) gltf_file: Option<String>,

    // activation state of Entity
    pub(crate) activation_state: ActivationState,

    // program local ID
    pub entity_id: u32,

    // component map
    pub(crate) components: TypeMap,

    #[cfg(debug_assertions)]
    component_names: Vec<String>,
}

unsafe impl Send for EntityObject {}
unsafe impl Sync for EntityObject {}

impl EntityObject {
    pub(crate) fn new(id: u32) -> Self {
        Self {
            #[cfg(debug_assertions)]
            debug_name: None,

            gltf_file: None,

            activation_state: ActivationState::default(),

            entity_id: id,

            components: TypeMap::default(),

            #[cfg(debug_assertions)]
            component_names: Vec::new(),
        }
    }

    pub fn gltf_file(&self) -> Option<&String> {
        self.gltf_file.as_ref()
    }

    pub fn multi_mut(&mut self) -> MultiMut<'_> {
        self.components.multi_mut()
    }

    pub fn insert_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        component: T,
    ) -> Option<T> {
        assert!(
            !self.activation_state.is_active(),
            "inserting components while the entity is activated is not allowed"
        );

        #[cfg(debug_assertions)]
        {
            let name = component.name().to_string();

            if !self.component_names.contains(&name) {
                self.component_names.push(name);
            }
        }

        self.components.insert(component)
    }

    pub(crate) fn insert_component_by_id(
        &mut self,
        type_id: TypeId,
        component: Box<dyn EntityComponent>,
    ) -> Option<Box<dyn EntityComponent>> {
        assert!(
            !self.activation_state.is_active(),
            "inserting components while the entity is activated is not allowed"
        );

        #[cfg(debug_assertions)]
        {
            let name = component.name().to_string();

            if !self.component_names.contains(&name) {
                self.component_names.push(name);
            }
        }

        self.components.insert_type(type_id, component)
    }

    pub fn remove_component<T: EntityComponent>(&mut self) -> Option<T> {
        assert!(
            !self.activation_state.is_active(),
            "removing components while the entity is activated is not allowed"
        );

        let t: Option<T> = self.components.remove();

        #[cfg(debug_assertions)]
        if let Some(t) = &t {
            if let Some(index) = self
                .component_names
                .iter()
                .position(|name| name == t.name())
            {
                self.component_names.remove(index);
            }
        }

        t
    }

    pub(crate) fn remove_component_by_id(
        &mut self,
        type_id: TypeId,
    ) -> Option<Box<dyn EntityComponent>> {
        assert!(
            !self.activation_state.is_active(),
            "removing components while the entity is activated is not allowed"
        );

        let t = self.components.remove_by_type_id(&type_id);

        #[cfg(debug_assertions)]
        if let Some(t) = &t {
            if let Some(index) = self
                .component_names
                .iter()
                .position(|name| name == t.name())
            {
                self.component_names.remove(index);
            }
        }

        t
    }

    #[cfg(debug_assertions)]
    pub fn component_names(&self) -> &[String] {
        &self.component_names
    }

    pub fn get_component<T: EntityComponent + ComponentDebug>(
        &self,
    ) -> std::result::Result<&T, ComponentNotFoundError> {
        self.components.get()
    }

    pub fn get_component_mut<T: EntityComponent + ComponentDebug>(
        &mut self,
    ) -> std::result::Result<&mut T, ComponentNotFoundError> {
        self.components.get_mut()
    }

    pub fn contains_component<T: EntityComponent>(&self) -> bool {
        self.components.contains::<T>()
    }

    pub fn is_activated(&self) -> bool {
        self.activation_state.is_active()
    }

    pub fn as_entity(&self) -> Entity {
        Entity { id: self.entity_id }
    }

    pub fn clone_without_components(&self, engine: &Engine) -> EntityObject {
        EntityObject {
            #[cfg(debug_assertions)]
            debug_name: self.debug_name.clone(),

            // gltf file name
            gltf_file: self.gltf_file.clone(),

            activation_state: Default::default(),

            entity_id: engine.assets().get_entity_id(),

            // component map
            components: TypeMap::new(),

            #[cfg(debug_assertions)]
            component_names: Vec::new(),
        }
    }

    pub fn clone_component_from<T: EntityComponent + ComponentDebug + Clone>(
        &mut self,
        other: &Self,
    ) -> Result<()> {
        let component = other.get_component::<T>()?;
        self.insert_component(component.clone());

        Ok(())
    }
}

#[allow(clippy::derive_hash_xor_eq)]
#[derive(Debug, Clone, Copy, Hash, Eq, Serialize, Deserialize)]
pub struct Entity {
    pub(crate) id: u32,
}

impl PartialEq<EntityObject> for Entity {
    fn eq(&self, entity: &EntityObject) -> bool {
        self.id == entity.entity_id
    }
}

impl PartialEq<Entity> for Entity {
    fn eq(&self, entity: &Entity) -> bool {
        self.id == entity.id
    }
}

impl fmt::Display for Entity {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.id)
    }
}

impl FromStr for Entity {
    type Err = ParseIntError;

    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
        Ok(Self { id: s.parse()? })
    }
}

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq, Serialize, Deserialize)]
pub struct NetworkID {
    pub(crate) id: u32,
}

impl NetworkID {
    pub fn display_vector(v: &Vec<Self>) -> String {
        // add starting bracket
        let mut s = String::from("[");

        for id in v {
            s += &format!("{},", id.id);
        }

        // remove last comma
        s.pop();

        // add closing bracket
        s += "]";

        s
    }

    pub fn from_opt(me: Option<Self>) -> String {
        match me {
            Some(nid) => format!("Some({})", nid),
            None => "None".to_string(),
        }
    }

    pub fn try_opt(s: &str) -> Result<Option<Self>> {
        Ok(if s == "None" {
            None
        } else {
            if !s.starts_with("Some") {
                panic!("unexpected string when parsing Option<NetworkID>: {}", s);
            }

            let tmp = s
                .trim_start_matches("Some")
                .trim_start_matches('(')
                .trim_end_matches(')');

            Some(tmp.parse()?)
        })
    }
}

impl fmt::Display for NetworkID {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.id)
    }
}

impl FromStr for NetworkID {
    type Err = ParseIntError;

    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
        Ok(Self { id: s.parse()? })
    }
}

impl Ord for NetworkID {
    fn cmp(&self, other: &Self) -> Ordering {
        self.id.cmp(&other.id)
    }
}

impl PartialOrd for NetworkID {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
