use crate::prelude::*;

use super::{
    directionallight::{DirectionalLight, ExtendedDirectionalLightData},
    pointlight::{ExtendedPointLightData, PointLight},
    spotlight::{ExtendedSpotLightData, SpotLight},
    RasterizerLightInterface,
};

use anyhow::Result;
use std::{
    marker::PhantomData,
    sync::{Arc, Mutex},
};
use utilities::prelude::cgmath::Vector3;

pub struct LightData<'a> {
    pub light_count: u32,
    pub info_descriptor: &'a Arc<DescriptorSet>,
    pub depth_map_descriptor: &'a Arc<DescriptorSet>,
}

pub struct LightHandler {
    pub image_count: usize,

    point_lights: LightWrapper<PointLight, ExtendedPointLightData>,
    directional_lights: LightWrapper<DirectionalLight, ExtendedDirectionalLightData>,
    spot_lights: LightWrapper<SpotLight, ExtendedSpotLightData>,
}

impl LightHandler {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        fallback: bool,
        shadow_image_size: u32,
    ) -> Result<Self> {
        Ok(Self {
            image_count,

            point_lights: LightWrapper::new(
                device,
                queue,
                image_count,
                fallback,
                shadow_image_size,
            )?,

            directional_lights: LightWrapper::new(
                device,
                queue,
                image_count,
                fallback,
                shadow_image_size,
            )?,

            spot_lights: LightWrapper::new(
                device,
                queue,
                image_count,
                fallback,
                shadow_image_size,
            )?,
        })
    }

    pub fn add(&mut self, light: Light) -> Result<()> {
        match light.light_type() {
            LightType::Point { .. } => self.point_lights.add(light),
            LightType::Direction { .. } => self.directional_lights.add(light),
            LightType::Spot { .. } => self.spot_lights.add(light),
        }
    }

    pub fn remove(&mut self, light: &Light) -> Result<bool> {
        match light.light_type() {
            LightType::Point { .. } => self.point_lights.remove(light),
            LightType::Direction { .. } => self.directional_lights.remove(light),
            LightType::Spot { .. } => self.spot_lights.remove(light),
        }
    }

    pub fn clear(&mut self) -> Result<bool> {
        let mut changed = false;

        if !self.point_lights.is_empty() {
            self.point_lights.clear()?;
            changed = true;
        }

        if !self.directional_lights.is_empty() {
            self.directional_lights.clear()?;
            changed = true;
        }

        if !self.spot_lights.is_empty() {
            self.spot_lights.clear()?;
            changed = true;
        }

        Ok(changed)
    }

    pub fn invalidate(&mut self) {
        self.point_lights.invalidate();
        self.directional_lights.invalidate();
        self.spot_lights.invalidate();
    }

    pub fn set_center(&mut self, center: Vector3<f32>) -> Result<()> {
        self.point_lights.set_center(center)?;
        self.directional_lights.set_center(center)?;
        self.spot_lights.set_center(center)?;

        Ok(())
    }

    pub fn render(
        &mut self,
        content: &Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        index: usize,
    ) -> Result<()> {
        self.point_lights.render(content, buffer_recorder, index)?;
        self.directional_lights
            .render(content, buffer_recorder, index)?;
        self.spot_lights.render(content, buffer_recorder, index)?;

        Ok(())
    }

    pub fn light_data(&self) -> (LightData<'_>, LightData<'_>, LightData<'_>) {
        (
            LightData {
                light_count: self.point_lights.lights.len() as u32,
                info_descriptor: &self.point_lights.info_descriptor,
                depth_map_descriptor: &self.point_lights.depth_map_descriptor,
            },
            LightData {
                light_count: self.directional_lights.lights.len() as u32,
                info_descriptor: &self.directional_lights.info_descriptor,
                depth_map_descriptor: &self.directional_lights.depth_map_descriptor,
            },
            LightData {
                light_count: self.spot_lights.lights.len() as u32,
                info_descriptor: &self.spot_lights.info_descriptor,
                depth_map_descriptor: &self.spot_lights.depth_map_descriptor,
            },
        )
    }

    pub fn light_info_descriptor_layout(
        device: &Arc<Device>,
        fallback: bool,
        count: u32,
    ) -> Result<Arc<DescriptorSetLayout>> {
        if fallback {
            DescriptorSetLayout::builder()
                .add_layout_binding(
                    0,
                    VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                    VK_SHADER_STAGE_FRAGMENT_BIT,
                    0,
                )
                .change_descriptor_count(count)
                .build(device.clone())
        } else {
            DescriptorSetLayout::builder()
                .add_layout_binding(
                    0,
                    VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                    VK_SHADER_STAGE_FRAGMENT_BIT,
                    VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT_EXT,
                )
                .change_descriptor_count(count)
                .build(device.clone())
        }
    }

    pub fn light_descriptor_layout(
        device: &Arc<Device>,
        fallback: bool,
        count: u32,
    ) -> Result<Arc<DescriptorSetLayout>> {
        if fallback {
            DescriptorSetLayout::builder()
                .add_layout_binding(
                    0,
                    VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                    VK_SHADER_STAGE_FRAGMENT_BIT,
                    0,
                )
                .change_descriptor_count(count)
                .build(device.clone())
        } else {
            DescriptorSetLayout::builder()
                .add_layout_binding(
                    0,
                    VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                    VK_SHADER_STAGE_FRAGMENT_BIT,
                    VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT_EXT,
                )
                .change_descriptor_count(count)
                .build(device.clone())
        }
    }
}

pub struct LightWrapper<T: RasterizerLightInterface<U>, U: ReprC + Send + Sync + 'static> {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,
    image_count: usize,
    shadow_image_size: u32,

    lights: Vec<T>,
    info_descriptor: Arc<DescriptorSet>,
    depth_map_descriptor: Arc<DescriptorSet>,

    light_buffer_type: PhantomData<U>,

    fallback: bool,
}

impl<T: RasterizerLightInterface<U>, U: ReprC + Send + Sync + 'static> LightWrapper<T, U> {
    fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        fallback: bool,
        shadow_image_size: u32,
    ) -> Result<Self> {
        let info_descriptor = Self::create_info_descriptor(device, fallback, 1)?;
        let depth_map_descriptor = Self::create_shadow_map_descriptor(device, fallback, 1)?;

        Ok(Self {
            device: device.clone(),
            queue: queue.clone(),
            image_count,
            shadow_image_size,

            lights: Vec::new(),
            info_descriptor,
            depth_map_descriptor,

            light_buffer_type: PhantomData,

            fallback,
        })
    }

    fn add(&mut self, light: Light) -> Result<()> {
        self.lights.push(T::new(
            light,
            &self.device,
            &self.queue,
            self.image_count,
            self.shadow_image_size,
        )?);
        self.update()?;

        Ok(())
    }

    fn remove(&mut self, light: &Light) -> Result<bool> {
        if let Some(i) = self.lights.iter().position(|pl| pl.eq(light)) {
            self.lights.remove(i);
            self.update()?;

            return Ok(true);
        }

        Ok(false)
    }

    fn is_empty(&self) -> bool {
        self.lights.is_empty()
    }

    fn clear(&mut self) -> Result<()> {
        self.lights.clear();
        self.update()
    }

    fn invalidate(&mut self) {
        for light in self.lights.iter_mut() {
            light.invalidate();
        }
    }

    fn set_center(&mut self, center: Vector3<f32>) -> Result<()> {
        for light in self.lights.iter_mut() {
            light.set_center(center)?;
        }

        Ok(())
    }

    fn render(
        &mut self,
        content: &Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        index: usize,
    ) -> Result<()> {
        for light in self.lights.iter_mut() {
            light.render(content, buffer_recorder, index)?;
        }

        Ok(())
    }

    fn update(&mut self) -> Result<()> {
        let buffers: Vec<&Arc<Buffer<U>>> = self.lights.iter().map(|l| l.buffer()).collect();

        if !buffers.is_empty() {
            self.info_descriptor =
                Self::create_info_descriptor(&self.device, self.fallback, buffers.len() as u32)?;

            self.info_descriptor
                .update(&[DescriptorWrite::uniform_buffers(0, &buffers)])?;
        }

        let mut textures = Vec::with_capacity(self.lights.len());

        if !textures.is_empty() {
            for light in &self.lights {
                let texture = light.shadow_map();
                Image::convert_layout(texture, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL)?;
                textures.push(texture);
            }

            self.depth_map_descriptor = Self::create_shadow_map_descriptor(
                &self.device,
                self.fallback,
                textures.len() as u32,
            )?;

            self.depth_map_descriptor
                .update(&[DescriptorWrite::combined_samplers(0, &textures)])?;
        }

        Ok(())
    }

    #[inline]
    fn create_info_descriptor(
        device: &Arc<Device>,
        fallback: bool,
        count: u32,
    ) -> Result<Arc<DescriptorSet>> {
        let descriptor_layout =
            LightHandler::light_info_descriptor_layout(device, fallback, count)?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(descriptor_layout)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool).allocate()
    }

    #[inline]
    fn create_shadow_map_descriptor(
        device: &Arc<Device>,
        fallback: bool,
        count: u32,
    ) -> Result<Arc<DescriptorSet>> {
        let descriptor_layout = LightHandler::light_descriptor_layout(device, fallback, count)?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(descriptor_layout)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool).allocate()
    }
}
