use crate::scene::content::events::Events;
use crate::scene::rendering::rasterizer::deferred::Rasterizer;
use crate::scene::rendering::rasterizer::traditional::TraditionalRasterizer;
use crate::scene::rendering::raytracer::hardwareraytracer::HardwareRayTracer;
use crate::scene::rendering::raytracer::raytracer::RayTracer;
use crate::scene::rendering::raytracer::softwareraytracer::SoftwareRayTracer;
use crate::scene::rendering::shared::shadercompiler::{ShaderCompiler, ShaderTypeConverter};
use crate::{prelude::*, scene::content::updates::Updates};

use crate::scene::content::entity::EntityObject;

use crate::scene::rendering::{ExtensionCheck, RenderingFrontEnd};

#[cfg(feature = "timings")]
use super::super::timings::Timings;

use super::scene_content::EntityNotFoundError;

use utilities::prelude::cgmath::{vec3, Matrix4, Vector3};

use anyhow::{bail, Error, Result};

use indexmap::IndexMap;
use rayon::prelude::*;

use std::any::{Any, TypeId};

use std::sync::Mutex;
use std::time::Instant;
#[cfg(feature = "timings")]
use std::time::Instant;

use std::{collections::HashSet, mem, time::Duration};
use std::{ptr::NonNull, sync::Arc};

pub struct EntityMultiMut<'a> {
    entities: &'a mut IndexMap<Entity, EntityObject>,
    buffer: HashSet<Entity>,
}

impl<'a> EntityMultiMut<'a> {
    pub(crate) fn new(entities: &'a mut IndexMap<Entity, EntityObject>) -> Self {
        Self {
            entities,
            buffer: HashSet::new(),
        }
    }

    pub fn get(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&'a mut EntityObject, EntityNotFoundError> {
        match self.entities.get_mut(&entity) {
            Some(v) => {
                let ptr = v as *mut EntityObject;

                assert!(
                    self.buffer.get(&entity).is_none(),
                    "Entity ({}) already borrowed",
                    entity.id
                );
                self.buffer.insert(entity);

                let e: Option<&'a mut EntityObject> = Some(unsafe { &mut *ptr });

                e.ok_or_else(|| EntityNotFoundError::new(entity))
            }
            None => Err(EntityNotFoundError::new(entity)),
        }
    }
}

pub struct Scene {
    screen_width: f32,
    screen_height: f32,

    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    pub(crate) render_type: SceneType,

    pub(crate) updates: Updates,
    pub(crate) archetypes: Archetypes,
    pub(crate) events: Events,
    pub resources: Resources,

    pub(crate) entities: IndexMap<Entity, EntityObject>,
    pub(crate) map: Option<Box<dyn Map>>,

    pub(crate) start_time: Instant,
    pub(crate) last_frame: Duration,
    pub(crate) frame_time: Duration,

    // renderer of this scene
    pub(crate) renderer: Box<dyn RenderingFrontEnd + Send + Sync>,

    pub(crate) frustum_check: Option<
        Box<
            dyn Fn(&EntityObject, &Matrix4<f32>, &ViewFrustum, &FrustumPlanes) -> Result<bool>
                + Send
                + Sync,
        >,
    >,

    pub(crate) particle_system_vulkan_objects: ParticleSystemVulkanObjects,

    #[cfg(feature = "timings")]
    pub(crate) timings: Timings,
}

impl Scene {
    pub(crate) fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        rasterizer_info: RasterizerInfo,
        raytracer_info: RaytracingInfo,
        graphics_info: GraphicsInfo,
    ) -> Result<Self> {
        let (renderer, render_type): (Box<dyn RenderingFrontEnd + Send + Sync>, SceneType) =
            Self::create_rendering_front_end(
                device,
                queue,
                (screen_width, screen_height),
                images,
                rasterizer_info,
                raytracer_info,
                graphics_info,
            )?;

        let compiler = ShaderCompiler::new(vec![(
            include_str!("../content/components/particle_system.comp"),
            "particle_system.comp",
        )])?;

        let cs = compiler.compile::<_, 0>(device, &[], |_| {})?;

        let particle_system_vulkan_objects = ParticleSystemVulkanObjects::new(device, queue, cs)?;

        let mut scene = Scene {
            screen_width,
            screen_height,

            render_type,

            device: device.clone(),
            queue: queue.clone(),

            updates: Updates::default(),
            archetypes: Archetypes::default(),
            events: Events::default(),
            resources: Resources::default(),
            entities: IndexMap::new(),
            map: None,

            frustum_check: None,

            renderer,

            start_time: Instant::now(),
            last_frame: Duration::from_secs(0),
            frame_time: Duration::from_secs(0),

            particle_system_vulkan_objects,

            #[cfg(feature = "timings")]
            timings: Timings::default()
                .add_timing("fire events")
                .add_timing("move events")
                .add_timing("entity changes")
                .add_timing("particles")
                .add_timing("frustum")
                .add_timing("renderer"),
        };

        scene.insert_default_updates()?;

        Ok(scene)
    }

    pub fn device(&self) -> &Arc<Device> {
        &self.device
    }

    pub fn queue(&self) -> &Arc<Mutex<Queue>> {
        &self.queue
    }
}

// object handling
impl Scene {
    #[inline]
    fn _entity_mut(
        entities: &mut IndexMap<Entity, EntityObject>,
        entity: Entity,
    ) -> Option<&mut EntityObject> {
        entities.get_mut(&entity)
    }

    pub fn set_map(&mut self, map_opt: Option<Box<dyn Map>>) -> Result<()> {
        if let Some(old_map) = self.map.take() {
            old_map.disable(self)?;
        }

        self.map = map_opt;

        Ok(())
    }

    pub fn clear_content(&mut self) -> Result<()> {
        let mut entities = IndexMap::new();
        mem::swap(&mut self.entities, &mut entities);

        for (_, entity) in entities.iter_mut() {
            for (_, component) in entity.components.iter_mut() {
                component.disable(self)?;
            }
        }

        self.map = None;
        self.updates.clear();
        self.archetypes.clear();
        self.events.clear();

        self.insert_default_updates()?;

        Ok(())
    }
}

impl Scene {
    pub fn add_archetype(&mut self, id: String, mut archetype: Archetype) -> Result<()> {
        archetype.setup(&self.entities)?;
        self.archetypes.insert(id, archetype);

        Ok(())
    }

    pub fn register_event<T: Any + Send + Sync>(&mut self) {
        self.events.register_event::<T>();
    }

    pub fn add_event_reader<T: Any + Send + Sync, F>(&mut self, f: F)
    where
        F: Fn(&mut SceneContents<'_>, &T) -> anyhow::Result<()> + Send + Sync + 'static,
    {
        self.events.add_reader(f);
    }

    pub fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.events.write_event(payload);
    }

    pub fn as_scene_contents(&mut self) -> SceneContents<'_> {
        SceneContents::new(
            self.render_type,
            self.frame_time,
            self.start_time,
            (self.screen_width, self.screen_height),
            &mut self.entities,
            &self.map,
            (&mut self.renderer, &mut self.particle_system_vulkan_objects),
            &mut self.events,
            &mut self.resources,
        )
    }

    pub(crate) fn update(&mut self) -> Result<()> {
        let mut temp_events = self.events.clone_from_register();

        let mut scene_content = SceneContents::new(
            self.render_type,
            self.frame_time,
            self.start_time,
            (self.screen_width, self.screen_height),
            &mut self.entities,
            &self.map,
            (&mut self.renderer, &mut self.particle_system_vulkan_objects),
            &mut temp_events,
            &mut self.resources,
        );

        #[cfg(feature = "timings")]
        let before = Instant::now();

        self.events.fire_events(&mut scene_content)?;

        #[cfg(feature = "timings")]
        self.timings
            .add("fire events", Instant::now().duration_since(before));

        #[cfg(feature = "timings")]
        let before = Instant::now();

        self.updates.update(&mut scene_content)?;

        #[cfg(feature = "timings")]
        self.timings
            .add("updates", Instant::now().duration_since(before));

        #[cfg(feature = "timings")]
        let before = Instant::now();

        self.events.move_events(scene_content.events.events);

        #[cfg(feature = "timings")]
        self.timings
            .add("move events", Instant::now().duration_since(before));

        #[cfg(feature = "timings")]
        let before = Instant::now();

        scene_content.entity_changes().commit(self)?;

        #[cfg(feature = "timings")]
        self.timings
            .add("entity changes", Instant::now().duration_since(before));

        Ok(())
    }

    pub fn execute_archetype(&mut self, name: impl ToString) -> Result<()> {
        let mut scene_content = SceneContents::new(
            self.render_type,
            self.frame_time,
            self.start_time,
            (self.screen_width, self.screen_height),
            &mut self.entities,
            &self.map,
            (&mut self.renderer, &mut self.particle_system_vulkan_objects),
            &mut self.events,
            &mut self.resources,
        );

        self.archetypes
            .execute(&name.to_string(), &mut scene_content)?;

        scene_content.entity_changes().commit(self)
    }

    pub fn archetype_info(&self, name: impl ToString) -> Result<ArchetypeInfo> {
        let name = name.to_string();
        let archetype = self
            .archetypes
            .get(&name)
            .ok_or_else(|| Error::msg(format!("requested archetype ({}) not found", name)))?;

        let mut entities = Vec::new();

        for entity in archetype.entities().keys() {
            #[cfg(debug_assertions)]
            let debug_name = self.entity(*entity)?.debug_name.clone();

            #[cfg(not(debug_assertions))]
            let debug_name = None;

            entities.push((*entity, debug_name));
        }

        Ok(ArchetypeInfo::new(entities))
    }
}

impl Scene {
    pub(crate) fn _world_to_screen_space(
        (screen_width, screen_height): (f32, f32),
        world_space: Vector3<f32>,
        view: &View,
    ) -> Result<(i32, i32)> {
        let (view, proj) = {
            let proj = view.proj();
            let view = view.camera().old_view_matrix();

            (view, proj)
        };

        let p = proj * view * world_space.extend(1.0);

        let pos = vec3(p.x / p.w, p.y / p.w, p.z / p.w);

        let x = Self::screen_space_to_coord(screen_width, pos.x);
        let y = Self::screen_space_to_coord(screen_height, pos.y);

        Ok((x, y))
    }

    pub(crate) fn insert_default_updates(&mut self) -> Result<()> {
        self.add_update(
            "animate",
            5_500_000,
            |scene_contents: &mut SceneContents<'_>,
             _entity,
             draw: &mut Draw,
             animation: &mut Animation| {
                animation.animate(scene_contents.now(), draw)?;

                Ok(())
            },
            EmptyFilter,
        )?;

        self.add_update(
            "particle_system",
            5_000_000,
            |scene_contents: &mut SceneContents<'_>,
             entity,
             particle_system: &mut ParticleSystem| {
                let now = scene_contents.now();

                if !particle_system.update(now) {
                    scene_contents.remove_entity(entity)?;
                    return Ok(());
                }

                {
                    let inheritance_info = CommandBuffer::inheritance_info(None, None, None, None);
                    let mut begin_info =
                        VkCommandBufferBeginInfo::new(VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT);
                    begin_info.set_inheritance_info(&inheritance_info);

                    let unsafe_scene = unsafe { remove_life_time_mut(scene_contents) };

                    let mut recorder = scene_contents
                        .particle_system_vulkan_objects_mut()
                        .begin(begin_info)?;

                    particle_system.animate(&mut recorder, unsafe_scene, now)?;
                }

                Ok(())
            },
            EmptyFilter,
        )?;

        Ok(())
    }

    pub fn set_frustum_check<F>(&mut self, f: F)
    where
        F: Fn(&EntityObject, &Matrix4<f32>, &ViewFrustum, &FrustumPlanes) -> Result<bool>
            + Send
            + Sync
            + 'static,
    {
        self.frustum_check = Some(Box::new(f));
    }

    fn create_rendering_front_end(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        rasterizer_info: RasterizerInfo,
        raytracer_info: RaytracingInfo,
        graphics_info: GraphicsInfo,
    ) -> Result<(Box<dyn RenderingFrontEnd + Send + Sync>, SceneType)> {
        let create =
            |render_type| -> Result<Option<(Box<dyn RenderingFrontEnd + Send + Sync>, SceneType)>> {
                match render_type {
                    SceneType::Rasterizer => {
                        if rasterizer_info.use_deferred {
                            match Rasterizer::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images,
                                graphics_info.render_scale,
                                rasterizer_info,
                                !Rasterizer::check(device),
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            };
                        } else {
                            match TraditionalRasterizer::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images,
                                rasterizer_info,
                                graphics_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            };
                        }
                    }
                    SceneType::Raytracer => {
                        if RayTracer::<HardwareRayTracer>::check(device) {
                            match RayTracer::<HardwareRayTracer>::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images.mono().len(),
                                graphics_info,
                                raytracer_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            }
                        }

                        if RayTracer::<SoftwareRayTracer>::check(device) {
                            match RayTracer::<SoftwareRayTracer>::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images.mono().len(),
                                graphics_info,
                                raytracer_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            }
                        };
                    }
                }

                Ok(None)
            };

        match create(graphics_info.render_type)? {
            Some(scene) => Ok(scene),
            None => {
                if let Some(scene_type) = Self::check_scene_type(
                    device,
                    graphics_info.render_type,
                    rasterizer_info.use_deferred,
                ) {
                    return Ok(create(scene_type)?.unwrap());
                }

                Err(anyhow::anyhow!("failed to create renderer"))
            }
        }
    }

    pub fn check_scene_type(
        device: &Arc<Device>,
        preferred_scene_type: SceneType,
        use_deferred: bool,
    ) -> Option<SceneType> {
        let check_rasterizer = |device, use_deferred| {
            if use_deferred {
                if Rasterizer::check(device) {
                    return Some(SceneType::Rasterizer);
                }
            } else {
                return Some(SceneType::Rasterizer);
            }

            None
        };

        match preferred_scene_type {
            SceneType::Rasterizer => check_rasterizer(device, use_deferred),
            SceneType::Raytracer => {
                if RayTracer::<HardwareRayTracer>::check(device)
                    || RayTracer::<SoftwareRayTracer>::check(device)
                {
                    return Some(SceneType::Raytracer);
                }

                check_rasterizer(device, use_deferred)
            }
        }
    }

    #[inline]
    pub(crate) fn screen_space_to_coord(max: f32, v: f32) -> i32 {
        (((v + 1.0) / 2.0) * max) as i32
    }

    pub fn view(&self) -> &View {
        self.renderer.view()
    }

    fn _process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> Result<()> {
        self.frame_time = {
            let now = self.now();

            let d = if self.last_frame < now {
                now - self.last_frame
            } else {
                Duration::from_secs(0)
            };

            self.last_frame = now;

            d
        };

        #[cfg(feature = "timings")]
        if let Some(timings) = self
            .timings
            .check_timing(self.now(), Duration::from_millis(2))
        {
            if !timings.is_empty() {
                println!("timings: \n{timings:#?}");
            }
        }

        // calculate movements and collision detection when map is present
        if self.map.is_some() {
            self.update()?;

            // queue particle system commands
            #[cfg(feature = "timings")]
            let before = Instant::now();

            self.particle_system_vulkan_objects
                .execute(buffer_recorder)?;

            #[cfg(feature = "timings")]
            self.timings
                .add("particles", Instant::now().duration_since(before));
        }

        #[cfg(feature = "timings")]
        let before = Instant::now();

        let content = match &self.frustum_check {
            Some(frustum_check) => {
                let view = self.renderer.view();
                let vp = view.proj() * view.camera().old_view_matrix();

                let frustum = view.frustum();
                let planes = frustum.planes();

                let content: Vec<Option<&EntityObject>> = self
                    .entities
                    .par_values()
                    .map(|entity_object| {
                        if frustum_check(entity_object, &vp, &frustum, &planes)? {
                            Ok(Some(entity_object))
                        } else {
                            Ok(None)
                        }
                    })
                    .collect::<Result<Vec<Option<&EntityObject>>>>()?;

                content.into_iter().flatten().collect()
            }
            None => self
                .entities
                .par_iter()
                .map(|(_entity, entity_object)| entity_object)
                .collect(),
        };

        #[cfg(feature = "timings")]
        self.timings
            .add("frustum", Instant::now().duration_since(before));

        #[cfg(feature = "timings")]
        let before = Instant::now();
        self.renderer
            .process(content, buffer_recorder, images, indices)?;

        #[cfg(feature = "timings")]
        self.timings
            .add("renderer", Instant::now().duration_since(before));

        Ok(())
    }
}

impl SceneEntities for Scene {
    fn entity(&self, entity: Entity) -> std::result::Result<&EntityObject, EntityNotFoundError> {
        self.entities
            .get(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))
    }

    fn entity_mut(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&mut EntityObject, EntityNotFoundError> {
        self.entities
            .get_mut(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))
    }

    fn entities_multi_mut(&mut self) -> EntityMultiMut<'_> {
        EntityMultiMut::new(&mut self.entities)
    }

    fn add_entity(&mut self, mut entity: EntityObject) -> Result<Entity> {
        // call enable event for all components
        for (_, component) in entity.components.iter_mut() {
            component.enable(self)?;
        }

        entity.activation_state.apply_change();

        let e = entity.as_entity();

        if entity.debug_name.clone() == Some("Lightning".to_string()) {
            println!(
                "Lightning added with components: {:#?}",
                entity.component_names()
            );
        }

        self.updates.add_entity(&mut entity, &self.entities)?;
        self.archetypes.add_entity(&mut entity)?;
        assert!(self.entities.insert(e, entity).is_none());

        Ok(e)
    }

    fn insert_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
        mut component: T,
    ) -> Result<()> {
        let entities = unsafe { remove_life_time_mut(&mut self.entities) };

        let entity_object = entities
            .get_mut(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))?;

        self.updates.remove_entity(entity);
        self.archetypes.remove_entity(entity);

        entity_object.activation_state.apply_change();

        component.enable(self)?;
        if entity_object.insert_component(component).is_some() {
            bail!("component {} already present", T::debug_name());
        }

        entity_object.activation_state.apply_change();

        self.updates.add_entity(entity_object, &self.entities)?;
        self.archetypes.add_entity(entity_object)?;

        Ok(())
    }

    fn remove_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
    ) -> Result<()> {
        let entities = unsafe { remove_life_time_mut(&mut self.entities) };

        let entity_object = entities
            .get_mut(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))?;

        self.updates.remove_entity(entity);
        self.archetypes.remove_entity(entity);

        entity_object.activation_state.apply_change();

        if let Some(mut component) = entity_object.remove_component_by_id(TypeId::of::<T>()) {
            component.disable(self)?;
        }

        entity_object.activation_state.apply_change();

        self.updates.add_entity(entity_object, &self.entities)?;
        self.archetypes.add_entity(entity_object)?;

        Ok(())
    }

    fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.write_event(payload);
    }

    fn remove_entity(&mut self, entity: Entity) -> Result<Option<EntityObject>> {
        if let Some(mut entity_object) = self.entities.swap_remove(&entity) {
            for (_, component) in entity_object.components.iter_mut() {
                component.disable(self)?;
            }

            entity_object.activation_state.apply_change();
            self.updates.remove_entity(entity);
            self.archetypes.remove_entity(entity);

            return Ok(Some(entity_object));
        }

        Ok(None)
    }

    // This function will panic if map is not set!
    fn map<F, T>(&mut self, f: F) -> Result<T>
    where
        F: FnOnce(&Box<dyn Map>, &mut SceneContents<'_>) -> Result<T>,
    {
        let map = self
            .map
            .as_ref()
            .expect("Map is not set while trying to call map");
        let p = map as *const Box<dyn Map> as *mut Box<dyn Map>;

        let mut scene_content = SceneContents::new(
            self.render_type,
            self.frame_time,
            self.start_time,
            (self.screen_width, self.screen_height),
            &mut self.entities,
            &self.map,
            (&mut self.renderer, &mut self.particle_system_vulkan_objects),
            &mut self.events,
            &mut self.resources,
        );

        let result = f(
            unsafe { NonNull::new_unchecked(p).as_ref() },
            &mut scene_content,
        )?;

        scene_content.entity_changes().commit(self)?;

        Ok(result)
    }

    fn now(&self) -> Duration {
        Instant::now().duration_since(self.start_time)
    }

    fn world_to_screen_space(&self, world_space: Vector3<f32>) -> Result<(i32, i32)> {
        Self::_world_to_screen_space(
            (self.screen_width, self.screen_height),
            world_space,
            self.view(),
        )
    }

    fn screen_space_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>> {
        let scale = self.renderer.render_scale();

        self.renderer
            .screen_to_world((x as f32 * scale) as u32, (y as f32 * scale) as u32)
    }

    fn particle_system_vulkan_objects_mut(&mut self) -> &mut ParticleSystemVulkanObjects {
        &mut self.particle_system_vulkan_objects
    }

    fn particle_system_vulkan_objects(&self) -> &ParticleSystemVulkanObjects {
        &self.particle_system_vulkan_objects
    }

    fn add_light(&mut self, light: Light) -> Result<()> {
        self.renderer.add_light(light)?;

        Ok(())
    }

    fn remove_light(&mut self, light: Light) -> Result<()> {
        self.renderer.remove_light(light)?;

        Ok(())
    }

    fn clear_lights(&mut self) -> Result<()> {
        self.renderer.clear_lights()?;

        Ok(())
    }

    fn view_mut(&mut self) -> &mut View {
        self.renderer.view_mut()
    }

    fn render_type(&self) -> SceneType {
        self.render_type
    }

    fn frame_time(&self) -> Duration {
        self.frame_time
    }
}

impl TScene for Scene {
    fn process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> anyhow::Result<()> {
        self._process(buffer_recorder, images, indices)
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
    ) -> anyhow::Result<()> {
        self.screen_width = window_width;
        self.screen_height = window_height;

        self.renderer.resize(window_width, window_height, images)
    }
}
