use crate::*;

use anyhow::Result;
use cgmath::Vector3;
use engine::prelude::*;
use entity_manager::*;
use rpg_components::{
    components::{level::Level, npc_type::NPCType},
    items::Loot,
};

use std::sync::Arc;

pub struct LootStash {
    pub loot: Vec<Option<Loot>>,

    action_radius: f32,
    open_chest_ui: Option<Arc<GuiBuilder>>,
    is_enabled: bool,
}

impl LootStash {
    pub const WORLD_OFFSET: f32 = 1.0;
    pub const MAX_STASH_SIZE: usize = 49;
}

impl LootStash {
    pub fn handle_npc_death(
        game: &Game,
        npc_level: &Level,
        npc_location: &Location,
        npc_type: NPCType,
    ) -> Result<Option<EntityObject>> {
        let loot_multiplier = match npc_type {
            NPCType::Boss => Some(game.item_settings.general.drop_chance_boss_multiplier),
            NPCType::Elite => Some(game.item_settings.general.drop_chance_elite_multiplier),
            NPCType::Normal => None,
        };

        let mut loot_list = vec![None; Self::MAX_STASH_SIZE];

        for loot_opt in loot_list.iter_mut() {
            match game
                .item_system()
                .get_random_loot(npc_level.level(), loot_multiplier)
            {
                Some(loot) => *loot_opt = Some(loot),
                None => break,
            }
        }

        if loot_list.iter().any(|loot_opt| loot_opt.is_some()) {
            return Ok(Some(LootStash::create(
                game,
                loot_list,
                npc_location.position(),
            )?));
        }

        Ok(None)
    }

    fn create(
        game: &Game,
        loot: Vec<Option<Loot>>,
        parent_position: Vector3<f32>,
    ) -> Result<EntityObject> {
        let mut loot_stash = game
            .entity_manager()
            .load_entity(game.engine().assets(), "lootchest")?;

        {
            let location = Location::new_and_setup(&mut loot_stash)?;
            location.set_position(parent_position);
            location
                .set_scale_uniform((loot.len() as f32 / (Self::MAX_STASH_SIZE * 2) as f32) + 0.5);

            loot_stash.insert_component(Self {
                loot,
                action_radius: game.game_settings().loot_action_radius,
                open_chest_ui: None,
                is_enabled: false,
            });

            loot_stash.insert_component(FactionFriendly);

            if let Ok(hitbox) = loot_stash.get_component_mut::<HitBox>() {
                hitbox.set_event(move |scene, _me, collider| -> Result<CollisionEventType> {
                    let collider_entity_object = scene.entity(collider)?;

                    if collider_entity_object
                        .get_component::<FactionPlayer>()
                        .is_ok()
                    {
                        return Ok(CollisionEventType::Block);
                    }

                    Ok(CollisionEventType::Ignore)
                })
            }
        }

        Ok(loot_stash)
    }

    pub fn is_empty(&self) -> bool {
        Self::is_list_empty(&self.loot)
    }

    pub fn remove_loot(&mut self, index: usize) -> Loot {
        self.loot[index].take().unwrap()
    }
}

impl LootStash {
    pub fn update(&self, (x, y): (i32, i32)) -> Result<()> {
        if let Some(ui) = &self.open_chest_ui {
            if self.is_enabled && !ui.visible() {
                ui.enable()?;
            }

            let grid: Arc<Grid> = ui.element("main grid")?;

            let (width, height) = grid.extents();
            grid.set_position(x - (width as i32 / 2), y - (height as i32 / 2))?;
        }

        Ok(())
    }

    pub fn enable_ui(&mut self, game: &Game) -> Result<()> {
        let open_chest_ui = {
            let ui = GuiBuilder::new(
                game.engine().gui_handler(),
                &game.build_data_path("gui/xml/ingame/open_loot.xml"),
            )?;

            let icon: Arc<Icon> = ui.element("open chest icon")?;

            match game
                .engine()
                .settings()
                .controller_button(game.engine(), ControllerButton::DPadUp)?
            {
                Some(pic) => {
                    icon.set_icon(&pic)?;
                }
                None => {
                    icon.set_icon(game.engine().settings().dark_key().as_ref().unwrap())?;
                    icon.set_text("E")?;
                }
            };

            let grid: Arc<Grid> = ui.element("main grid")?;

            // this is super important to prevent weird behavior
            grid.disallow_position_scale()?;

            ui
        };

        debug_assert!(self.open_chest_ui.is_none());
        self.open_chest_ui = Some(open_chest_ui);

        Ok(())
    }

    pub fn disable_ui(&mut self) -> Result<()> {
        debug_assert!(self.open_chest_ui.is_some());

        if self.is_enabled {
            self.open_chest_ui.as_ref().unwrap().disable()?;
        }

        self.open_chest_ui = None;

        Ok(())
    }

    pub fn create_debug(
        game_handle: &GameHandle,
        loot: Vec<Option<Loot>>,
        parent_position: Vector3<f32>,
    ) -> Result<EntityObject> {
        let game = game_handle.upgrade();

        let mut loot_stash = game
            .entity_manager()
            .load_entity(game.engine().assets(), "lootchest")?;

        {
            let location = Location::new_and_setup(&mut loot_stash)?;
            location.set_position(parent_position);
            location
                .set_scale_uniform((loot.len() as f32 / (Self::MAX_STASH_SIZE * 2) as f32) + 0.5);

            let me = Self {
                loot,
                action_radius: game.game_settings().loot_action_radius,
                open_chest_ui: None,
                is_enabled: false,
            };

            loot_stash.insert_component(me);
            loot_stash.insert_component(FactionFriendly);
        }

        Ok(loot_stash)
    }
}

impl LootStash {
    pub fn is_list_empty(loot: &[Option<Loot>]) -> bool {
        let mut is_empty = true;

        for loot_opt in loot.iter() {
            if loot_opt.is_some() {
                is_empty = false;
                break;
            }
        }

        is_empty
    }

    pub fn radius(&self) -> f32 {
        self.action_radius
    }
}

impl EntityComponent for LootStash {
    fn enable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.is_enabled = true;

        if let Some(ui) = &self.open_chest_ui {
            debug_assert!(!ui.visible());

            ui.enable()?;
        }

        Ok(())
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.is_enabled = false;

        if let Some(ui) = &self.open_chest_ui {
            debug_assert!(ui.visible());

            ui.disable()?;
        }

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for LootStash {
    fn debug_name() -> &'static str {
        "LootStash"
    }
}
