use syn::{
    parse::{Parse, ParseStream, Result},
    Attribute, Ident, Token,
};

#[derive(Debug)]
pub enum AttributeTokenType {
    PreSetterExec(Ident),
    SetterExec(Ident),
    ApplyExec(Ident),
    ConvertAs(proc_macro2::TokenStream),
    SkipGetter,
    PrivateSetter,
    CustomFrom(Ident),
    CustomInto(Ident),
    CompileFeature(proc_macro2::TokenStream),
    CompileNotFeature(proc_macro2::TokenStream),
}

impl Parse for AttributeTokenType {
    fn parse(input: ParseStream) -> Result<Self> {
        let _attrs = input.call(Attribute::parse_outer)?;
        let attribute_type = input.parse::<Ident>()?.to_string();

        match attribute_type.as_str() {
            "pre_setter_exec" => {
                let _equal: Token![=] = input.parse()?;
                let function_name: Ident = input.parse()?;

                Ok(Self::PreSetterExec(function_name))
            }
            "setter_exec" => {
                let _equal: Token![=] = input.parse()?;
                let function_name: Ident = input.parse()?;

                Ok(Self::SetterExec(function_name))
            }
            "apply_exec" => {
                let _equal: Token![=] = input.parse()?;
                let function_name: Ident = input.parse()?;

                Ok(Self::ApplyExec(function_name))
            }
            "convert_as" => {
                let _equal: Token![=] = input.parse()?;
                let data_type: proc_macro2::TokenStream = input.parse()?;

                Ok(Self::ConvertAs(data_type))
            }
            "skip_getter" => Ok(Self::SkipGetter),
            "private_setter" => Ok(Self::PrivateSetter),
            "from" => {
                let _equal: Token![=] = input.parse()?;
                let function_name: Ident = input.parse()?;

                Ok(Self::CustomFrom(function_name))
            }
            "into" => {
                let _equal: Token![=] = input.parse()?;
                let function_name: Ident = input.parse()?;

                Ok(Self::CustomInto(function_name))
            }
            "feature" => {
                let _equal: Token![=] = input.parse()?;
                let feature_name: proc_macro2::TokenStream = input.parse()?;

                Ok(Self::CompileFeature(feature_name))
            }
            "not_feature" => {
                let _equal: Token![=] = input.parse()?;
                let feature_name: proc_macro2::TokenStream = input.parse()?;

                Ok(Self::CompileNotFeature(feature_name))
            }

            _ => panic!("unknown Token: {}", attribute_type),
        }
    }
}

impl PartialEq for AttributeTokenType {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (Self::PreSetterExec(_), Self::PreSetterExec(_)) => true,
            (Self::SetterExec(_), Self::SetterExec(_)) => true,
            (Self::ApplyExec(_), Self::ApplyExec(_)) => true,
            (Self::ConvertAs(_), Self::ConvertAs(_)) => true,
            (Self::SkipGetter, Self::SkipGetter) => true,
            (Self::PrivateSetter, Self::PrivateSetter) => true,
            (Self::CustomFrom(_), Self::CustomFrom(_)) => true,
            (Self::CustomInto(_), Self::CustomInto(_)) => true,
            (Self::CompileFeature(_), Self::CompileFeature(_)) => true,
            (Self::CompileNotFeature(_), Self::CompileNotFeature(_)) => true,

            _ => false,
        }
    }
}
